/*
	Tracemac
	Traces a MAC address through a switched Ethernet network

	Copyright (C) 2003 Brady Alleman, All Rights Reserved.
	
	Uses NetMRG as a foundation.
*/

// Include system headers
#include <list>
#include <string>
#include <cstdio>
#include <cstdlib>
#include <iostream>

using namespace std;

// Include the NetMRG Headers
#include "types.h"
#include "utils.h"
#include "snmp.h"

bool debug = false;
string vlan = "";
int max_hop_count = 20;
int hop_count = 0;

list<string> parse_communities(string commlist)
{
	int pos;
	list<string> communities;
	while ((pos = commlist.find(",")) != string::npos)
	{
		communities.push_back(commlist.substr(0,pos));
		commlist.erase(0, pos+1);
	}
	communities.push_back(commlist);
	return communities;
}

string erase_quotes(string astr)
{
	int pos;
	while ((pos = astr.find("\"")) != string::npos)
	{
		astr.erase(pos, 1);
	}
	return astr;
}

void print_list(list<string> alist)
{
	for (list<string>::iterator cur = alist.begin(); cur != alist.end(); cur++)
	{
		printf("'%s'\n", cur->c_str());
	}
}

string find_working_community(string ip, list<string> communities)
{
	DeviceInfo info;
	info.ip = ip;
	info.snmp_version = 1;
	info.snmp_timeout = 500000;
	info.snmp_retries = 1;
	long long int uptime;
	
	for (list<string>::iterator cur = communities.begin(); cur != communities.end(); cur++)
	{
		info.snmp_read_community = *cur;
		snmp_session_init(info);
		uptime = get_snmp_uptime(info);
		snmp_session_cleanup(info);
		if (uptime) return *cur;
	}
	
	return "";
}

string clean_mac(const string & mac)
{
	string temp = "";
	for (int i = 0; i < mac.length(); i++)
	{
		if (isxdigit(mac[i]))
		{
			temp += tolower(mac[i]);
		}
	}
	return temp;	
}

string parse_delim_mac(const string & mac, int delims)
{
	string blocks[delims+1];

	int curblock = 0;
	for (int i = 0; i < mac.length(); i++)
	{
		if (isxdigit(mac[i]))
		{
			blocks[curblock] += tolower(mac[i]);
		}
		else
		{
			curblock++;
		}
	}

	string result = "";
	
	for (curblock = 0; curblock <= delims; curblock++)
	{
		for (int k = blocks[curblock].length(); k < (12 / (delims + 1)); k++)
		{
			blocks[curblock] = "0" + blocks[curblock];
		}
		result += blocks[curblock];
	}

	return result;
}

string parse_user_mac(const string & mac)
{
	int delim_count = 0;
	for (int i = 0; i < mac.length(); i++)
	{
		if (!isxdigit(mac[i]))
		{
			delim_count++;
		}
	}

	switch (delim_count)
	{
		case 5:		
		case 3:		
		case 2:		
		case 1:		return parse_delim_mac(mac, delim_count);
		default:	return clean_mac(mac);
	}
}

string pretty_mac(const string & mac)
{
	string temp = mac;
	string res;
	while (!temp.empty())
	{
		res.append(1, (char) toupper(temp[0]));
		res.append(1, (char) toupper(temp[1]));
		res += ":";
		temp.erase(0,2);
	}
	res.erase(res.length()-1,1);
	return res;
}

string hex_to_dec(const string & hex)
{
	unsigned int value;
	sscanf(hex.c_str(), "%x", &value);
	return inttostr(value);
}

string cdpaddress_to_ip(const string & cdpip)
{
	unsigned int a, b, c, d;
	sscanf(cdpip.c_str(), "%x %x %x %x", &a, &b, &c, &d);
	return inttostr(a) + "." + inttostr(b) + "." + inttostr(c) + "." + inttostr(d);
}

string get_cdp_index(DeviceInfo & info, string portifindex)
{
	list<SNMPPair> cdpAddressType = snmp_walk(info, ".1.3.6.1.4.1.9.9.23.1.2.1.1.3." + portifindex);
	cdpAddressType = snmp_trim_rootoid(cdpAddressType, ".1.3.6.1.4.1.9.9.23.1.2.1.1.3.");
	if (!cdpAddressType.empty())
		return cdpAddressType.begin()->oid;
	return "U";
}

string find_mac_index(DeviceInfo & info, string mactofind)
{
	/*
	list<SNMPPair> bridgeAddressList = snmp_walk(info, ".1.3.6.1.2.1.17.4.3.1.1");
	bridgeAddressList = snmp_trim_rootoid(bridgeAddressList, ".1.3.6.1.2.1.17.4.3.1.1.");
	for (list<SNMPPair>::iterator cur = bridgeAddressList.begin(); cur != bridgeAddressList.end(); cur++)
	{
		if (clean_mac(cur->value) == mactofind)
		{
			return cur->oid;
		}
	}
	return "";
	*/
	return 	hex_to_dec(mactofind.substr(0,2)) + "." +
		hex_to_dec(mactofind.substr(2,2)) + "." +
		hex_to_dec(mactofind.substr(4,2)) + "." +
		hex_to_dec(mactofind.substr(6,2)) + "." +
		hex_to_dec(mactofind.substr(8,2)) + "." +
		hex_to_dec(mactofind.substr(10,2));
}

int num_mac_addrs(DeviceInfo & info, string bridge_port)
{
	int count = 0;
	list<SNMPPair> ports = snmp_walk(info, ".1.3.6.1.2.1.17.4.3.1.2");
	ports = snmp_trim_rootoid(ports, ".1.3.6.1.2.1.17.4.3.1.2.");
	for (list<SNMPPair>::iterator cur = ports.begin(); cur != ports.end(); cur++)
	{
		if (cur->value == bridge_port)
		{
			if (strtoint(snmp_get(info, ".1.3.6.1.2.1.17.4.3.1.3." + cur->oid)) == 3)
				count++;
		}
	}
	return count;
}

string find_next_node(string node, string mactofind, list<string> communities)
{
	string next_hop = "";
	string community = find_working_community(node, communities);
	if (community == "")
	{
		printf("No working community for node %s, terminating.\n", node.c_str());
	}
	else
	{
		if (debug)
			printf("Node %s accepts community '%s'.\n", node.c_str(), community.c_str());

		if ((vlan != "") && (vlan != "0") && (vlan != "U") && (vlan != "1"))
			community += "@" + vlan;

		if (debug)
			printf("Community with VLAN: %s\n", community.c_str());
	
		DeviceInfo info;
		info.ip = node;
		info.snmp_version = 1;
		info.snmp_timeout = 850000;
		info.snmp_retries = 2;
		info.snmp_read_community = community;
		snmp_session_init(info);

		string bridge_index = find_mac_index(info, mactofind);
		string bridge_port = snmp_get(info, ".1.3.6.1.2.1.17.4.3.1.2." + bridge_index);
		if (bridge_port == "U")
		{
			printf("MAC address not known to this switch, terminating.\n");
		}
		else
		{
			string port_ifindex = snmp_get(info, ".1.3.6.1.2.1.17.1.4.1.2." + bridge_port);
			string port_name = snmp_get(info, "ifDescr." + port_ifindex);
			if (port_name.length() > 21)
				port_name = snmp_get(info, "ifName." + port_ifindex);

			if (debug)
			{
				int mac_count = num_mac_addrs(info, bridge_port);
				printf("MAC address is on bridge index #%s, port #%s, ifindex #%s, port name: %s, number of MACs: %d", bridge_index.c_str(), bridge_port.c_str(), port_ifindex.c_str(), port_name.c_str(), mac_count);
			}
			else
			{
				printf("%-17s ---> %-21s", node.c_str(), port_name.c_str());
			}

			string cdp_index = get_cdp_index(info, port_ifindex);

			if (cdp_index == "U")
			{
				// check for PortChannel via CISCO-PAGP-MIB
				list<SNMPPair> port_channel_assoc = snmp_walk(info, ".1.3.6.1.4.1.9.9.98.1.1.1.1.8");
				port_channel_assoc = snmp_trim_rootoid(port_channel_assoc, ".1.3.6.1.4.1.9.9.98.1.1.1.1.8.");
				list<string> port_channel_members;
				for (list<SNMPPair>::iterator pc = port_channel_assoc.begin(); pc != port_channel_assoc.end(); pc++)
				{
					if (pc->value == port_ifindex)
					{
						// we are a member of this port group
						port_channel_members.push_back(snmp_get(info, "ifName." + pc->oid));
						if (cdp_index == "U")
						{
							cdp_index = get_cdp_index(info, pc->oid);
						}
					}
				}

				if (cdp_index != "U")
				{
					port_channel_members.reverse();
					printf(" { ");
					for (list<string>::iterator amember = port_channel_members.begin(); amember != port_channel_members.end(); amember++)
					{
						if (amember != port_channel_members.begin())
							printf(", ");
						printf("%s", amember->c_str());
					}
					printf(" } ");
				}
			}

			if (cdp_index == "U")
			{
				// check for PortChannel via IEEE8023ad MIB
				list<SNMPPair> port_channel_assoc = snmp_walk(info, ".1.2.840.10006.300.43.1.2.1.1.13");
				port_channel_assoc = snmp_trim_rootoid(port_channel_assoc, ".1.2.840.10006.300.43.1.2.1.1.13.");
				list<string> port_channel_members;
				for (list<SNMPPair>::iterator pc = port_channel_assoc.begin(); pc != port_channel_assoc.end(); pc++)
				{
					if (pc->value == port_ifindex)
					{
						// we are a member of this port group
						port_channel_members.push_back(snmp_get(info, "ifName." + pc->oid));
						if (cdp_index == "U")
						{
							cdp_index = get_cdp_index(info, pc->oid);
						}
					}
				}

				if (cdp_index != "U")
				{
					port_channel_members.reverse();
					printf(" { ");
					for (list<string>::iterator amember = port_channel_members.begin(); amember != port_channel_members.end(); amember++)
					{
						if (amember != port_channel_members.begin())
							printf(", ");
						printf("%s", amember->c_str());
					}
					printf(" } ");
				}
			}

			if (cdp_index == "U")
			{
				int mac_count = num_mac_addrs(info, bridge_port);
				if (debug)
					printf("No CDP information for this port.\n");
				else
					printf(" ---> ");
				if (mac_count == 1)
					printf("Target device (%s).\n", pretty_mac(mactofind).c_str());
				else
				if (mac_count == 0)
					printf("Nothing on this port?\n");
				else
					printf("Switch or hub without CDP.  (%d MACs)\n", mac_count);
			}
			else
			{
				printf("\n");
				string cdp_port = erase_quotes(snmp_get(info, ".1.3.6.1.4.1.9.9.23.1.2.1.1.7." + cdp_index));
				string raw_cdp_ip = snmp_get(info, ".1.3.6.1.4.1.9.9.23.1.2.1.1.4." + cdp_index);
				string cdp_ip = cdpaddress_to_ip(raw_cdp_ip);
				string cdp_vlan = snmp_get(info, ".1.3.6.1.4.1.9.9.23.1.2.1.1.11." + cdp_index);
				//vlan = cdp_vlan;
				string tag = "";

				if (node == cdp_ip)
				{
					//printf("Looping?\n");

					/* CDP says next hop is on the same IP address
					   This Means:
						1) We're clustered 
						-or-
						2) Something evil is happening 

					   For Cisco Catalyst Clustering:
						1) Member Switches are accessible using community <read-community>@esN (N = Sw#)
						2a) Replies to these types of requests (other than the master/es0) are NATed by the cisco
						2b) The NAT isn't clean, the source ports of the replies are not 161
						2c) Use of this behind a firewall will die, as these 'wrong' packets won't be allowed back in
						3) The Cisco specs say that a member switch need not have an IP address, SNMP only via master
						3a) As my current employer doesn't use this, this code ignores that possibility for now
						4) I've probably not anticipated lots of things going wrong here, your mileage may vary
						5) Patches welcome!
					*/
					

					if (snmp_get(info, ".1.3.6.1.4.1.9.9.134.1.1.10.0") != "U")
					{
						//printf("Clustering?\n");

						// this node appears to be a cluster master
						
						string cdp_sys_name = erase_quotes(snmp_get(info, ".1.3.6.1.4.1.9.9.23.1.2.1.1.6." + cdp_index));
						
						string oldcomm = info.snmp_read_community;
						string member_num = "U";
							
						list<SNMPPair> memberNums = snmp_walk(info, ".1.3.6.1.4.1.9.9.134.1.2.1.1.2");
						for (list<SNMPPair>::iterator cur = memberNums.begin(); cur != memberNums.end(); cur++)
						{
							info.snmp_read_community = oldcomm + "@es" + cur->value;
							snmp_session_cleanup(info);
							snmp_session_init(info);
							string snmp_sys_name = snmp_get(info, ".1.3.6.1.2.1.1.5.0");
							//printf("Comparing '%s' to '%s'\n", cdp_sys_name.c_str(), snmp_sys_name.c_str());
							if (snmp_sys_name == cdp_sys_name)
								member_num = cur->value;
						}

						if (member_num == "U")
						{
							printf("Clustering: CDP device name failed to match name of any cluster members.\n");
							info.snmp_read_community = oldcomm;
						}
						else
						{
							info.snmp_read_community = oldcomm + "@es" + member_num;
							snmp_session_cleanup(info);
							snmp_session_init(info);
							list<SNMPPair> ipAddrs = snmp_walk(info, ".1.3.6.1.2.1.4.20.1.1");
							if (!ipAddrs.empty())
							{
								cdp_ip = ipAddrs.begin()->value;
								tag = "C";
							}
							else
							{
								printf("Clustering: Failed to find any IPs on cluster member.\n");
							}
						}

						


					}

					
				}


				if (debug)
				{
					printf("Next hop found via CDP: %s [%s] - %s [%s] - VLAN %s\n\n", cdp_ip.c_str(), raw_cdp_ip.c_str(), cdp_port.c_str(), cdp_index.c_str(), cdp_vlan.c_str());
				}
				else
				{
					printf("%-17s <--- %-21s\n", string(cdp_ip + " " + tag).c_str(), cdp_port.c_str());
				}

				next_hop = cdp_ip;
			}
		}
		snmp_session_cleanup(info);
	}
	return next_hop;
}

void show_version()
{
	printf("Tracemac\n");
	printf("Version 0.02\n");
	printf("\n");
}

void show_usage()
{
	show_version();
	printf("-h                    - help\n");
	printf("-V                    - version\n");
	printf("-c comm1,comm2,comm3  - comma separated list of snmp community strings (required)\n");
	printf("-r ip                 - first switch to probe (root node, required)\n");
	printf("-v vlan id            - vlan id for root node\n");
	printf("-m xx:xx:xx:xx:xx:xx  - MAC address to find (required)\n");
	printf("-x count              - Maximum hop count\n");
	printf("-d                    - debugging info\n");
	printf("-a                    - all debugging info\n");
	printf("\n");
}

int main(int argc, char **argv)
{
	int option_char;
	string commlist, rootnode, mactofind;

	if (argc < 3)
	{
		show_usage();
		exit(0);
	}

	while ((option_char = getopt(argc, argv, "hVdac:r:m:v:x:")) != EOF)
		switch (option_char)
		{
			case 'h': 	show_usage();
						exit(0); 
						break;
			case 'V':	show_version();
						exit(0);
						break;
			case 'd':	debug = true;
						break;
			case 'c':	commlist = optarg;
						break;
			case 'r':	rootnode = optarg;
						break;
			case 'v':	vlan = optarg;
						break;
			case 'm':	mactofind = parse_user_mac(optarg);
						break;
			case 'a':	set_debug_level(LEVEL_ALL);
						set_debug_components(DEBUG_ALL);
						break;
			case 'x':	max_hop_count = atoi(optarg);
						break;
		}
		
	list<string> communities = parse_communities(commlist);
	
	if (debug)
	{
		printf("SNMP communities to search:\n");
		print_list(communities);
		printf("\n");
		printf("Trying to find %s starting at %s.\n\n", pretty_mac(mactofind).c_str(), rootnode.c_str());
	}
	
	snmp_init();
	
	string node = rootnode;
	string lastnode = rootnode;
	while ( (node != "") && (hop_count < max_hop_count) )
	{
		node = find_next_node(node, mactofind, communities);
		hop_count++;
		if (node == lastnode)
		{
			printf("Switch points to itself?\n");
			break;
		}
		lastnode = node;
	}
	if (hop_count == max_hop_count)
	{
		printf("Maximum hop count (%d) exceeded.  Bridging loop?\n", max_hop_count);
	}

	snmp_cleanup();
	
	return 0;
}
